# -*- coding: utf-8 -*- 
# ------------------------------------------------------------------------------
# Name:          wordCollector.py
# Purpose:       WordCollector class.
#
#                WordCollector creates a MarkovChain
#                from a word stream (strings). The corresponding
#                Producer class is SentenceProducer which reverses the process.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright � 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.collector import Collector
from common.textParser import TextParser
import pandas as pd

import re

class WordCollector(Collector):
    """Create a Markov chain of words from a body of text.
    
    WordCollector analyzes word collections (Strings delimited by white space and/or punctuation) 
    and collects statistics in the form of a MarkovChain that can be used by a Producer class.
    
    For a given string (inline or from a file/stream), this examines all the word collections
    of a given length 1 to n (n <= 5), and records the #instances of each word that
    follows that word collection. Sentences are formed left to right advancing 1 Word each iteration.
    
    Features used in the corresponding SentenceProducer class:
    Sentence ending punctuation (. ! ?) is retained as a word to indicate the end of a sentence.
    A prefix (default is a space) is prepended to the first word of each sentence/line to indicate an initial word.
    
    There are three collection modes: words, sentences, and lines.
    The collection mode determines the set of initial keys. An initial key indicates the
    start of some unit - a sentence or line (which may have > 1 sentence).
    In 'words' collection mode, there are no initial keys - all the words treated equally.
    

    """
    
    def __init__(self, state_size=2, verbose=0, source=None, text=None, maxlines=None, ignore_case=True, remove_stop_words=False):
        super().__init__(state_size, verbose, source,  domain='text')
        self.text = text
        self.processing_mode = 'sentences'         # set by command line arguments to 'words' or 'sentences'
        
        self._text_parser = None
        self.ignore_case = ignore_case
        self._maxlines = maxlines
        self._remove_stop_words = remove_stop_words
        self._terminal_object = '~'
        self._initial_object = ' '
        self._source = source
        self._words = None                  # words in order of appearance
        
        self.countsFileName = '_wordCounts' + '_0{}'.format(state_size)
        self.chainFileName = '_wordsChain' + '_0{}'.format(state_size)
        
        self.words_re = re.compile(r'[,;: ]')    # regular expression to split a line into words

    def __str__(self):
        return f"WordCollector order={self.order} verbose={self.verbose} name={self.name} format={self.format}, source={self.source}, text={self.text}, ignoreCase={self.ignoreCase}"


    def _process_words(self, words:list) -> pd.DataFrame:
        """Process a list of words to create a counts_df DataFrame
        
        The algorithm is identical to that used in CharacterCollector
        the difference being this operates on words instead of individual characters.
        counts_df DataFrame has 3 columns: 'key', 'word', and 'count' and a range index.
        key is a string of n-words where n is the order (state space size) of the MarkovChain model
        Initial keys are prefixed by TextParser._prefix, which defaults to a space.
        
        """
        word_keys =  [ (' '.join(words[i:i+self.order]), words[i+self.order]) for i in range(0, len(words)-2) ]
        word_keys_df =  pd.DataFrame(data=word_keys, columns=['key','word'])
        words_ser = word_keys_df.value_counts(ascending=False)
        df = pd.DataFrame(words_ser, columns=['count'])
        self.counts_df = df.reset_index()
        if self.sort_chain:
            self.counts_df.sort_values(by=['key','word'], ignore_index=True, inplace=True)
            
        return self.counts_df

    def _set_keys(self):
        prefix = self._text_parser.initial_word_prefix
        self._initial_keys_df = self.counts_df[ [x.startswith(prefix) for x in self.counts_df['key']] ]
        #
        # initiial_keys and keys are both sets and so the values are unique
        #
        self._initial_keys = set(self._initial_keys_df['key'].values.tolist())
        self._keys = set(self.counts_df['key'].values.tolist())
        if self.verbose > 0:
            print(f'total number of keys: {len(self._keys)}')
            print(f'number of initial keys: {len(self._initial_keys)}')
    
    def collect(self):
        self._text_parser = TextParser(txt=self.text, source=self._source, \
                   maxlines=self._maxlines, ignore_case=self.ignore_case, \
                   remove_stop_words=self._remove_stop_words)
        #
        # word counts are informational only
        #
        self.word_counts = self._text_parser.get_word_counts(sort_counts=True, reverse=True)
        self.words_df = self._text_parser.counts_df
                
        if self.processing_mode == 'words':
            self._collect_words()
        elif self.processing_mode == 'sentences':
            self._collect_sentences()
        elif self.processing_mode == 'lines':
            self._collect_lines()
        
        # create the MarkovChain from the counts by summing probabilities
        if self.counts_df.size > 0:
            self._set_keys()
            self._create_chain(initial=True)
            

    def _collect_lines(self) -> pd.DataFrame:
        """Run collection on a list of lines  using the set parameters
        
        Returns: MarkovChain result
        """
        #
        # get the words in the order they appear in the text
        #
        self._words = self._text_parser.get_line_words()
        return self._process_words(self._words)
        
            
    def _collect_sentences(self) -> pd.DataFrame:
        """Run collection on a list of sentences using the set parameters
        
        Returns: MarkovChain result
        """
        #
        # get the words in the order they appear in the text
        #
        self._words = self._text_parser.get_sentence_words()
        return self._process_words(self._words)
        

    def _collect_words(self) -> pd.DataFrame:
        """Run collection on a list of words
        
        Returns: MarkovChain result
        """

        #
        # get the words in the order they appear in the text
        #
        if self._remove_stop_words:
            self._words = self._text_parser.get_words()
        else:
            self._words = self._text_parser.get_all_words()
        
        #
        # create a counts_df DataFrame
        #
        return self._process_words(self._words)


    def save(self):
        save_result = super().save()
        return save_result

if __name__ == '__main__':
    print(WordCollector.__doc__())
        



from music21 import interval, instrument, corpus, converter
from music21 import note, clef, duration, metadata, key, pitch, scale
from music21.stream import Score
from music21.stream import Part
from music21.stream import Measure
from music.instruments import Instruments

import re
import pandas as pd
import pathlib, random, copy, math
from datetime import date
from builtins import isinstance
import matplotlib.pyplot as plt
import seaborn as sns

class MusicUtils(object):
    """Music utilities
        TODO - the method names are confusing, like score_notes vs. notes_for_score etc. Need to refactor
    """
    
    verbose = 0
    C_Major = key.Key('C')
    A_Minor = key.Key('a')
    default_pitch_map = {'Db':'C#', "D-":'C#', 'D#':'Eb', 'Gb':'F#', 'G-':'F#', 'Ab':'G#', 'A-':'G#', 'A#':'Bb' }
     
    @staticmethod
    def get_score(file_path:str) -> Score:
        score = None
        file_info = MusicUtils.get_file_info(file_path)
        if file_info['exists']:
                score = converter.parse(file_info['path_text'])
        return score
    
    @staticmethod
    def get_score_parts_info(ascore, partnames=None, display=False):
        part_dict = {}
        for apart in ascore.parts:
            if partnames is None or (partnames is not None and apart.partName in partnames):
                ksigs, mnumbers = MusicUtils.get_keySignatures(apart)
                for ks in ksigs:
                    if not isinstance(ks, key.Key):
                        ks = ks.asKey()
                    if display:
                        print(f'  {apart.partName}  {ks} {ks.mode}')
                    part_dict[apart.partName] = {'key':ks, 'keyName':ks.name, 'mode':ks.mode }
        return part_dict
    
    @staticmethod
    def get_parts_info(scores, partnames=None, display=False):
        all_scores = {}
        for ascore in scores:
            title = ascore.metadata.title
            if display: print(title)
            all_scores[title] = MusicUtils.get_score_parts_info(ascore, partnames, display)
        return all_scores
    
    #
    # get intervals for a Part
    #
    @staticmethod
    def get_part_intervals(apart:Part) -> [dict]:
        """Get all intervals for a Part.
        
        Arguments:
            apart - a Part instance
        Returns:
            A list of intervals where each interval is a dict with the keys "note1", "note2"
            the integer interval (number of semitones) needed to go from note1 to note2.
        """
        intrvals = []
        part_notes = apart.flat.getElementsByClass('Note')
        for ind in range(len(part_notes)-1):
            n1 = part_notes[ind]
            n2 = part_notes[ind+1]
            i = interval.Interval(n1, n2)
            idict = {'note1':n1, 'note2':n2, 'interval':i}
            intrvals.append(idict)
        return intrvals
    
    @staticmethod
    def get_part_notes(apart:Part) -> [note.Note]:
        part_notes = apart.flat.getElementsByClass('Note')
        return part_notes
    
    @staticmethod
    def get_scale_degree(akey:key.Key, anoteorpitch:object) -> dict:
        modifier = ''
        apitch = anoteorpitch
        if isinstance(anoteorpitch, note.Note):
            apitch = anoteorpitch.pitch
        sd,accidental = akey.getScaleDegreeAndAccidentalFromPitch(apitch)
        if accidental is not None:
            modifier = accidental.modifier
        sdstr = f'{modifier}{sd}'
        roman = f'{modifier}{akey.romanNumeral(sd).figure}'
        
        return {'pitch':apitch, 'pitch_name':apitch.name, 'number':sdstr, 'roman':roman}

    @staticmethod
    def get_pitch_from_scaleDegree(akey:scale.ConcreteScale, sd:str, octave=None) -> pitch.Pitch:
        """Given a scale degree and a Key, return a deepcopy of the appropriate pitch
        
        The music21 pitchFromDegree api only handles integer scale degrees
        This accepts a string with the scale degree prepended with an accidental #, - or b
        so for example, in Key('C'), "-3" is actually an e- (e-flat)
        "#4" and "-5" are both a F4.
        NOTE that the range of pitches returned is C4 to B4. The octave can be set
        using the optional octave argument.
        """

        p = None
        if sd.isdigit():
            p = akey.pitchFromDegree(int(sd))
        elif len(sd) >= 2:
            accidental = sd[0]
            semitones = 1
            if accidental == '-' or accidental == 'b':
                semitones = -1
            degree = int(sd[1:])
            p = akey.pitchFromDegree(degree)
            p = p.transpose(interval.Interval(semitones))
        if octave is not None:
            p.octave = octave
        return p
    
    @staticmethod
    def get_scale_degrees(apart:Part) -> [dict]:
        """Gets the scale degree of every note's pitch in a given Part
        
        Arguments:
            apart - a music21.stream.Part instance
        Returns:
            a list of dictionary entries for each Part note.
            The keys are 'pitch', 'pitch_name', 'number' and 'roman' (for a Roman figure).
            where 'pitch' is a Pitch object, 'number' and 'roman' are scale degrees as a string.
            Any alterations are indicated by prepending a # or - for sharp, flat.
            An empty list is returned if there is no Key or KeySignature associated with the Part
            
        Note that Chords are not handled. In particular, notes that are part of a Chord are ignored.
        """
        key_sigs,measure_numbers = MusicUtils.get_keySignatures(apart)
        measure_numbers.append(len(apart))
        scale_degrees = []
        for i in range(len(key_sigs)):
            k = key_sigs[i]
            if not isinstance(k, key.Key):
                k = k.asKey()
            partMeasures = apart.measures(measure_numbers[i], measure_numbers[i+1]-1)   # measure range is inclusive
            pNotes = partMeasures.flat.getElementsByClass('Note')
            partPitches = pNotes.pitches
            for p in partPitches:
                scale_degrees.append(MusicUtils.get_scale_degree(k, p))
        return scale_degrees

    @staticmethod
    def get_score_intervals(ascore:Score, partname:str=None) -> dict:
        """ Get the intervals for all Parts of a Score
        
        Arguments:
            ascore - a Score instance
            partname - a part name in the Score or None to get intervals for all parts
        Returns:
            A dictionary of intervals where the key is the part name. The key values
            is a list of intervals where each interval is a dict with the keys "note1", "note2"
            the integer interval (number of semitones) needed to go from note1 to note2.
        """
        parts = ascore.getElementsByClass(Part)
        pdict = dict()
        for p in parts:
            pname = p.partName
            if partname is None or pname==partname:
                pdict[pname] = MusicUtils.get_part_intervals(p)
        return pdict
    
    @staticmethod
    def get_score_notes(ascore:Score, partname:str=None) -> dict:
        """Get the Notes for all Parts or the named part of a Score as a dict with the part name as the key, and a [note.Note] as the value.
        
        Note that this does not return Rest or Chord objects
        """
        parts = ascore.getElementsByClass(Part)
        pdict = dict()
        for p in parts:
            pname = p.partName
            if partname is None or pname==partname:
                notes = MusicUtils.get_part_notes(p)
                pdict[pname] = notes
        return pdict

    @staticmethod
    def get_music21_objects_for_score(classinfo, ascore:Score, partnames:[str]=None, partnumbers:[int]=None) ->  (pd.DataFrame,{str},{int}):
        if classinfo is note.Note:
            return MusicUtils.get_notes_for_score(ascore, partnames, partnumbers)
        elif classinfo is interval.Interval:
            return MusicUtils.get_intervals_for_score(ascore, partnames, partnumbers)
        else:
            raise TypeError
    
    @staticmethod
    def get_intervals_for_score(ascore:Score, partnames:[str]=None, partnumbers:[int]=None) ->  (pd.DataFrame,{str},{int}):
        """Get the intervals of specified Parts of a Score as a pandas.DataFrame
        Arguments:
            ascore - a Score instance
            partnames - a list of part names to extract from Score. If None then intervals from all parts are extracted.
        
        DataFrame columns returned:
            interval  (interval.Interval)
            part_name
            part_number
            note1  (note.Note the first Note in the interval pair)
            note2  (note.Note the second Note in the interval pair)
            name (interval name as in "P5")
            directedName  (as in "P-5" for down a 5th)
            niceName
            semitones (int)
        
        Returns a 3-tuplet consisting of the intervals_df DataFrame,
        a [int] of part numbers, and a [str] of part names.
        """
        pdict = MusicUtils.get_score_intervals(ascore)
        intrvals_df = pd.DataFrame()
        part_number = 1
        df = None
        score_partnames = set()
        score_partnumbers = set()
        if partnames is None:
            partnames = list(pdict.keys())
        have_partnames = not ( partnames is None or len(partnames)==0)
        have_partnumbers = not (partnumbers is None or len(partnumbers)==0 or part_number in partnumbers)
        for k in pdict.keys():  # part names
                
                if (have_partnames and k in partnames) or (have_partnumbers and part_number in partnumbers):
                    part_intervals = pdict[k]
                    intervals =  [x['interval'] for x in part_intervals]
                    df = pd.DataFrame(data=intervals, columns=['interval'])
                    df['part_number'] = part_number
                    df['part_name'] = k
                    df['note1'] = [x['note1'] for x in part_intervals]
                    df['note2'] = [x['note2'] for x in part_intervals]
                    score_partnames.add(k)
                    score_partnumbers.add(part_number)
                    intrvals_df = pd.concat([intrvals_df, df])
                    part_number = part_number + 1
        if df is not None:
            # else this score has none of the parts specified 
            intrvals_df['name'] = [x.name for x in intrvals_df['interval']]
            intrvals_df['directedName'] = [x.directedName for x in intrvals_df['interval']]
            intrvals_df['niceName'] = [x.niceName for x in intrvals_df['interval']]
            intrvals_df['semitones'] = [x.semitones for x in intrvals_df['interval']]
            intrvals_df.reset_index(drop=True, inplace=True)       # make sure index is unique
        return intrvals_df, score_partnames, score_partnumbers
    

    @staticmethod
    def get_notes_for_score(ascore:Score, partnames:[str]=None, partnumbers:[int]=None) ->  (pd.DataFrame,{str},{int}):
        """Get the Notes of specified Parts from a score as a pandas.DataFrame
        Arguments:
            ascore - a Score instance
            partnames - a list of part names to extract from Score. If None, no parts are extracted.
                        If ['*'] (the default), then notes from all parts are extracted.
                                
        DataFrame columns returned:
            note (music21.note.Note)
            part_number
            part_name
            name  (just the pitch name as in "D")
            nameWithOctave   (pitch name with the octave as in "D5")
            pitch (music21.pitch.Pitch)
            pitchClass (integer)
            ps (pitch space - float)
            duration (music21.duration.Duration)
            type (the duration type string, for example "quarter")
            ordinal
            quarterLength (duration.quarterLength, float)
            quarterLengthNoTuplets
            scaleDegree (the ordinal of the pitch relative to the part's Key/KeySignature)
            
        Returns a 3-tuple consisting of the notes_df DataFrame,
        an integer set of part numbers, and a set(str) of part names.
        """
        pdict = MusicUtils.get_score_notes(ascore)    # {partname1 : [note.Note], partname2 : [note.Note], ... }
        score_parts = MusicUtils.get_score_parts(ascore)
        notes_df = pd.DataFrame()
        scale_degrees = []
        df = None
        part_number = 1
        score_partnames = set()
        score_partnumbers = set()
        if partnames is None:
            partnames = list(pdict.keys())
        have_partnames = not ( partnames is None or len(partnames)==0)
        have_partnumbers = not (partnumbers is None or len(partnumbers)==0 or part_number in partnumbers)
        for k in pdict.keys():
            if  (have_partnames and k in partnames) or (have_partnumbers and part_number in partnumbers):
                df = pd.DataFrame(data=pdict[k], columns=['note'])
                df['part_number'] = part_number
                df['part_name'] = k
                sdlist = MusicUtils.get_scale_degrees(score_parts[k])  # scale degrees for notes in this Part
                scale_degrees.extend([x['number'] for x in  sdlist])
                score_partnames.add(k)
                score_partnumbers.add(part_number)
                notes_df = pd.concat([notes_df, df])
                part_number = part_number + 1
                
        if df is not None:
            # else this score has none of the parts specified 
            dfnotes = notes_df['note']
            notes_df['name'] = [x.name for x in dfnotes]
            notes_df['nameWithOctave'] = [x.nameWithOctave for x in dfnotes]
            notes_df['pitch'] = [x.pitch for x in dfnotes]
            notes_df['pitchClass'] = [x.pitch.pitchClass for x in dfnotes]
            notes_df['ps'] = [x.pitch.ps for x in dfnotes]
            notes_df['duration'] = [x.duration for x in dfnotes]
            notes_df['type'] = [x.duration.type for x in dfnotes]
            notes_df['ordinal'] = [x.duration.ordinal for x in dfnotes]
            notes_df['quarterLength'] = [x.duration.quarterLength for x in dfnotes]
            notes_df['quarterLengthNoTuplets'] = [x.duration.quarterLengthNoTuplets for x in dfnotes]
            notes_df['scaleDegree'] = scale_degrees
            
            notes_df.reset_index(drop=True, inplace=True)       # make sure index is unique
        return notes_df, score_partnames, score_partnumbers
    
    @staticmethod
    def get_durations_from_notes(source_df:pd.DataFrame) -> pd.DataFrame:

        """Get the Durations from a notes or intervals DataFrame
        
        The notes_df argument has the columns described in get_notes_for_score().
        DataFrame columns returned:
            note (music21.note.Note)
            duration (music21.duration.Duration)
            type
            ordinal
            dots
            fullName
            quarterLength
            tuplets
        See the music21.duration documentation for details on individual fields
        """
        if 'duration' in source_df.columns:
            notes_df = source_df
            durations_df = pd.DataFrame(data=notes_df[['note','duration']], columns=['note','duration'])
            durations_df['type'] = [x.type for x in notes_df['duration']]
            durations_df['ordinal'] = [x.ordinal for x in notes_df['duration']]
            durations_df['dots'] = [x.dots for x in notes_df['duration']]
            durations_df['fullName'] = [x.fullName for x in notes_df['duration']]
            durations_df['quarterLength'] = [x.quarterLengthNoTuplets for x in notes_df['duration']]
            durations_df['tuplets'] = [x.tuplets for x in notes_df['duration']]
        else:
            intervals_df = source_df
            durations_df = pd.DataFrame(data=intervals_df[['note1']], columns=['note1'])
            durations_df['duration'] = [x.duration for x in intervals_df['note1']]
            durations_df['type'] = [x.duration.type for x in intervals_df['note1']]
            durations_df['ordinal'] = [x.duration.ordinal for x in intervals_df['note1']]
            durations_df['dots'] = [x.duration.dots for x in intervals_df['note1']]
            durations_df['fullName'] = [x.duration.fullName for x in intervals_df['note1']]
            durations_df['quarterLength'] = [x.duration.quarterLengthNoTuplets for x in intervals_df['note1']]
            durations_df['tuplets'] = [x.duration.tuplets for x in intervals_df['note1']]
            durations_df.rename(columns={'note1':'note'}, inplace=True)
        return durations_df
    
    @staticmethod
    def get_metadata_bundle(composer:str=None, title:str=None) -> metadata.bundles.MetadataBundle:
        """Gets a MetadataBundle for a composer and title.
            Arguments:
                composer - composer name, can be None
                title - a regular expression string used to do a title search.
                
            If title is just a string (no re characters) it will do an exact match.
            For example, to search for all Bach's works (in the corpus) that starts with "bwv1"
            specify get_metadata_bundle(composer='bach', title='^bwv1+')
        
        """
        meta = None
        if title is not None:
            titles_re = re.compile(title, re.IGNORECASE)
        if composer is not None:
            meta = corpus.search(composer,'composer')
            if title is not None:
                meta = meta.intersection(corpus.search(titles_re,'title'))
        elif title is not None:
            meta = corpus.search(titles_re,'title')
        return meta
    
    @staticmethod
    def get_all_score_music21_objects(classinfo, composer=None, title=None, partnames=None, partnumbers=None):
        if classinfo is note.Note:
            return MusicUtils.get_all_score_notes(composer, title, partnames, partnumbers)
        elif classinfo is interval.Interval:
            return MusicUtils.get_all_score_intervals(composer, title, partnames, partnumbers)
        else:
            raise TypeError

    @staticmethod
    def get_all_score_notes(composer=None, title=None, partnames=None, partnumbers=None):
        scores, titles = MusicUtils.get_scores_from_corpus(composer, title)
        notes_df = pd.DataFrame()
        all_score_partnames = set()
        all_score_partnumbers = set()
        for i in range(len(scores)):
            score = scores[i]
            df,pnames,pnums = MusicUtils.get_notes_for_score(score, partnames, partnumbers)
            df['title'] = titles[i]
            notes_df = pd.concat([notes_df, df], ignore_index=True)
            all_score_partnames = all_score_partnames.union(pnames)
            all_score_partnumbers = all_score_partnumbers.union(pnums)
        return notes_df, all_score_partnames, all_score_partnumbers
            
    @staticmethod
    def get_all_score_intervals(composer=None, title=None, partnames=None, partnumbers=None):
        meta = MusicUtils.get_metadata_bundle(composer, title)
        intrvals_df = pd.DataFrame()
        all_score_partnames = set()
        all_score_partnumbers = set()
        for i in range(len(meta)):
            md = meta[i].metadata
            if MusicUtils.verbose > 0:
                    print(f"working on {md.title}")
            score = corpus.parse(meta[i])
            df,pnames,pnums = MusicUtils.get_intervals_for_score(score, partnames, partnumbers)
            if len(df) > 0:
                df['title'] = md.title
                intrvals_df = pd.concat([intrvals_df, df], ignore_index=True)
                all_score_partnames = all_score_partnames.union(pnames)
                all_score_partnumbers = all_score_partnumbers.union(pnums)
        return intrvals_df, all_score_partnames, all_score_partnumbers
    
    @staticmethod
    def get_scores_from_corpus(composer=None, title=None, keypart:str=None, filters:dict=None):
        """Gets a list of Score objects from the corpus matching the composer and title provided.
        
        Args:
            composer = the name of the composer (as it appears in the corpus), for example 'bach'
            title = the title to search. This can be a regular expression, for example '^bwv31[0-3]', or a single work name as in 'bwv310'
            keypart - the name of the Part (if any) that determines the overall score Key.
            filters - a dictionary of filter values. Currently only filter is 'mode' which can be 'major' or 'minor'
        Returns:
            a 2-element tupple of the list of matching Score, and a list of corresponding str (titles)
            
        Note this uses MusicUtils.get_metadata_bundle which uses music21.corpus to do the search.
        Because the filter is at the Score level, the 'mode' filter (if present) is applied only when the keypart argument is not None.
        If mode filter and keypart are both provided, then the Key to all Parts is set to be the same as in keypart.
        
        Note - if keypart is provided but that part is NOT in a score, that score is NOT added to the lists of scores & titles returned.
        """
        scores = []
        titles = []
        meta = MusicUtils.get_metadata_bundle(composer, title)
        for i in range(len(meta)):
            md = meta[i].metadata
            ascore = corpus.parse(meta[i])

            if filters is not None and 'mode' in filters and keypart is not None:     # check if the score matches the filter
                mode = filters['mode']
                spinfo = MusicUtils.get_score_parts_info(ascore, partnames=[keypart])
                if keypart in spinfo and spinfo[keypart]['mode']==mode:
                    if MusicUtils.verbose > 0:
                        print(f'title: {md.title}')
                    #
                    # update the key to all parts
                    #
                    new_key = None
                    if 'key' in spinfo[keypart]:
                        new_key = spinfo[keypart]['key']
                    for apart in ascore.parts:
                        if apart.partName != keypart:
                            measures = apart.getElementsByClass(Measure)
                            for m in measures:
                                keys = m.getElementsByClass([key.Key, key.KeySignature])
                                if len(keys) > 0:
                                    for k in keys:
                                        if MusicUtils.verbose > 1:
                                            print(f'update {m} in {md.title}  {apart.partName} to key:{new_key}')
                                        if new_key is not None:
                                            m.removeByClass([key.Key, key.KeySignature])
                                            m.insert(spinfo[keypart]['key'])
                    scores.append(ascore)
                    titles.append(md.title)
            else:   
                scores.append(ascore)
                titles.append(md.title)
                
        return scores,titles
    
    @staticmethod
    def show_measures(measures, how='text'):
        nmeasures = len(measures)
        print(f"number of measures: {nmeasures}")
        i = 1
        for measure in measures:
            notes = measure.getElementsByClass(['Note','Chord','Rest'])
            print('measure {}'.format(i))
            for n in notes:
                if n.isNote:
                    print(n.nameWithOctave)
                else:
                    print(n.fullName)
            i = i+1
            measure.show(how)   # if None, displays in MuseScore
    
    @staticmethod
    def show_intervals(df, what='name') -> str:
        int_string = ''
        for ki in df[what].values.tolist():
            int_string = int_string + str(ki) + ','
        return int_string.rstrip(',')
    
    @staticmethod
    def show_notes(df, what='name') -> str:
        """Creates a stringified list of notes from a DataFrame
            Args:
                df - the source DataFrame
                what - the DataFrame column to select ('name', 'nameWithOctave' or 'scaleDegree')
            For example, show_notes(notes_df, 'nameWithOctave'), len(notes_df)==2, would return
            a string like 'F#4,C#5'
        
        """
        notes_str = ''
        for kn in df[what].values.tolist():
            notes_str = notes_str + kn + ','
        return (notes_str.rstrip(','))
    
    @staticmethod
    def show_durations(df, what='quarterLength') -> str:
        duration_string = ''
        for ds in df[what].values.tolist():
            duration_string = duration_string + str(ds) + ','
        return duration_string.rstrip(',')
    
    @staticmethod
    def to_string(alist:[int]) -> str:
        """Stringifies an integer list.
            For example [-2,3] --> "-2,3", because of str(alist) formatting
            str([-2,3]) --> "['2', '3']"
            
        """
        lstr = ''
        for x in alist:
            lstr = lstr + str(x) + ','
        return lstr.rstrip(',')

    @staticmethod
    def note_info(note):
        dur = note.duration
        if note.isRest:
            info = f'name: {note.name}, fullName: {note.fullName}, type: {dur.type}, dots: {dur.dots}, quarterLength: {dur.quarterLength}'
        else:
            info = f'{note.nameWithOctave}, type: {dur.type}, dots: {dur.dots}, fullName: {dur.fullName}, quarterLength: {dur.quarterLength}, tuplets: {dur.tuplets}'
        return info

    @staticmethod
    def duration_info(dur) -> duration.Duration:
        info = f'type: {dur.type}, ordinal: {dur.ordinal}, dots: {dur.dots}, fullName: {dur.fullName}, \
        quarterLength {dur.quarterLength}, quarterLengthNoTuplets {dur.quarterLengthNoTuplets}, tuplets: {dur.tuplets}'
        return info
    
    @staticmethod
    def get_interval_stats(ascore, partnames=None, partnumbers=None):
        int_df,pnames,pnums = MusicUtils.get_intervals_for_score(ascore, partnames, partnumbers)
        int_df = int_df.groupby(by=['semitones']).count()[['interval']]
        int_df.rename(columns={'interval':'count'}, inplace=True)
        int_df.reset_index(inplace=True)       
        return int_df.sort_values(by='count', ascending=False)

    @staticmethod
    def create_note(start_note:note.Note, 
                    anInterval:interval.Interval=None, semitones:int=None,
                    dur:duration.Duration=None, quarterLen:int=None) -> note.Note:
        """Creates a new Note a given interval away from a starting note
        
        Args:
            start_note (note.Note): the anchor Note
            interval (interval.Interval): a valid Interval or None. If None, semitones must be present.
            semitones (int): interval expressed as semitones
            duration (duration.Duration): duration to assign to the new Note
            quarterLength (int): duration to assign to the new Note expressed as quarterLength
        Returns:
            A new Note that is the specified interval away from the start_note.
        Notes:
            * duration or quarterLength may be specified, but not both.
            * if both duration and quarterLength are none, the duration of the new Note
              will be the same as the duration of start_note
        """
        new_note = copy.copy(start_note)    # shallow copy
        if anInterval is None and semitones is None:
            raise ValueError('interval and semitones cannot both be None')
        intval = anInterval
        if anInterval is None:
            intval = interval.Interval(semitones)
        new_note = new_note.transpose(intval)
        if dur is None and quarterLen is not None:
            new_note.quarterLength = quarterLen
        elif dur is not None:
            new_note.duration = dur
        return new_note
    
    @staticmethod
    def random_notes(lower_ps:int, upper_ps:int, num:int=10, minval=0.5) -> [note.Note]:
        notes = []
        for n in range(num):
            ps = random.randrange(lower_ps,upper_ps)    # random pitch (ps values)
            dur = random.randrange(1, 8) * minval       # random durations as quarterLength
            notes.append(note.Note(ps=ps, quarterLength=dur))
        return notes
    
    @staticmethod
    def random_part(lower_ps:int, upper_ps:int, num:int=10, minval=0.5, 
                     clef=clef.TrebleClef(), instrument=instrument.Piano()):
        # minval is given as quarterLength, 0.5 is an eighth note, 0.25 is a 16th etc.
        part = Part()
        part.insert(0,instrument)
        notes = []
        notes.append(clef)
        for n in range(num):
            ps = random.randrange(lower_ps,upper_ps)    # random pitch (ps value) c4 to b6
            dur = random.randrange(1, 8) * minval    # random durations (multiples of 16th notes) as quarterLength
            notes.append(note.Note(ps=ps, quarterLength=dur))
        part.append(notes)
        return part
    
    @staticmethod
    def random_score(partnames:[str] = ['Soprano','Alto','Tenor','Bass'], length:int=20, min_duration:float=0.5, scale_name=None) -> Score:
        score_instruments = []
        score = Score()
        instruments_info = Instruments()
        instruments_pd = instruments_info.instruments_pd
        for pn in partnames:
            part_instrument = instruments_pd.loc[pn].instance
            score_instruments.append(part_instrument)
            pitch_range = instruments_pd.loc[pn]['range_ps']
            clef_name = instruments_pd.loc[pn]['clef']
            clef = instruments_info.clefs_pd.loc[clef_name]['instance']
            part = MusicUtils.random_part(pitch_range[0], pitch_range[1], length, min_duration, clef, instrument=part_instrument)
            score.append(part)
        return score
    
    @staticmethod
    def get_file_info(cpath:str, def_extension:str='mxl') -> dict:
        """Breaks up a path name into component parts
        
        """
        
        known_extensions = [def_extension, 'mxl','.xml','.musicxml','.json','.txt','.csv']
        x = cpath.split("/")
        paths = x[0:len(x)-1]
        filename = x[-1]
        ext = filename.split(".")
        name = ext[0]
        if len(ext)==2 and ext[1] in known_extensions:
            ext = ext[1]
            path = cpath      
        else:
            ext = def_extension
            filename = f"{filename}.{ext}"
            path = f"{cpath}.{ext}"
        p = pathlib.Path(path)
        exists = p.exists()
        return  {'paths':paths, 'path_text':path, 'filename':filename, 'name':name,'extension': ext, 'Path':p, 'exists':exists}

    @staticmethod
    def round_values(x, places=5):
        if not type(x) is str:
            return round(x, places)
        else:
            return x
    
    @staticmethod
    def get_timedelta(start_date=date(1970,1,1), end_date=date.today()) -> int:
        """Gets the number of elapsed seconds between two dates.
        
        Args:
            start_date (datetime.date): the starting date. Default value is Jan 1,1970
            end_date: the ending date. Default value is today()
        Returns:
            The absolute value of the number of elapsed seconds between the two dates.
        Notes:
            Useful in setting a random seed.
        
        """
        
        delta = end_date - start_date
        return abs(int(delta.total_seconds()))
    
    @staticmethod
    def get_score_parts(aScore:Score):
        """Get the Parts of a Score
        Args:
            aScore - a stream.Score instance
        Returns:
            A dict of stream.Part with the key partName
        
        """
        parts_dict = dict()
        if aScore is not None:
            parts = aScore.getElementsByClass(Part)
            for p in parts:
                parts_dict[p.partName] = p
        return parts_dict
    
    @staticmethod
    def get_keySignatures(apart:Part, default_keySignature:key.KeySignature=None):
        """Gets the KeySignatures appearing in a given Part
        
            Args:
                apart : a stream.Part instance
                default_keySignature : use this if not None for all KeySignatures/Keys
            Returns:
                a tupple consisting of 2 lists:
                [KeySignature] in the order of appearance
                [int] of measure numbers where the KeySignature starts
        """
        measures = apart.getElementsByClass(Measure)
        key_sigs = []
        measure_numbers = []
        for m in measures:
            keys = m.getElementsByClass([key.Key,key.KeySignature])
            for k in keys:
                if k is not None:
                    if default_keySignature is not None:
                        key_sigs.append(default_keySignature)
                    else:
                        key_sigs.append(k)
                    measure_numbers.append(m.measureNumber)
        return key_sigs,measure_numbers
    
    @staticmethod
    def get_score_keySignatures(ascore:Score) -> {str:tuple}:
        """Gets the KeySignatures of all parts of a Score
            Args:
                ascore - a Score instance
            Returns:
                a dictionary of tupple, where the key is the part name and the tuple is ([KeySignature], [measures])
        """
        parts = ascore.getElementsByClass(Part)
        pdict = dict()
        for apart in parts:
            pdict[apart.partName] = MusicUtils.get_keySignatures(apart)
        return pdict

    @staticmethod
    def adjust_accidental(aNote:note.Note, preference:object=default_pitch_map, inPlace=True) -> note.Note:
        adjusted_note = aNote
        if not inPlace:
            adjusted_note = copy.deepcopy(aNote)
        if isinstance(preference, dict):
            keys = preference.keys()

        p = adjusted_note.pitch.simplifyEnharmonic()
        if p.accidental is None:
            accidental = 'natural'
        else:
            accidental = p.accidental.name
        if isinstance(preference, dict):
            if p.name in keys:
                p = p.getEnharmonic()
        elif preference=='sharp' and 'flat' in accidental:
                p = p.getEnharmonic()
        elif preference=='flat'  and 'sharp' in accidental:
                p = p.getEnharmonic()
        adjusted_note.pitch = p
        
        return adjusted_note
                
    @staticmethod
    def adjust_accidentals(apart: Part, preference:object=default_pitch_map, inPlace=True) -> Part:
        """Adjusts the accidentals on all notes in a given Part.
            Arguments:
                apart - a music21.stream.Part instance which can be the entire part, or measures()
                preference - "sharp", "flat" or dict() of pitch mappings. Default is default_pitch_map:
                             {'Db':'C#', "D-":'C#', 'D#':'Eb', 'Gb':'F#', 'G-':'F#', 'Ab':'G#', 'A-':'G#', 'A#':'Bb' }
                inPlace - if True changes are applied to the Part argument
            Returns:
                The adjusted Part
        """
        adjusted_part = apart
        if not inPlace:
            adjusted_part = copy.deepcopy(apart)
        part_notes = MusicUtils.get_part_notes(adjusted_part)

        for aNote in part_notes:
            aNote = MusicUtils.adjust_accidental(aNote, preference, inPlace=True)
        return adjusted_part

    @staticmethod
    def adjust_score_accidentals(ascore:Score, preference:object=default_pitch_map, partnames:[str]=None, inPlace=True) -> Score:
        """Adjusts the accidentals on all notes in a given score for specified part names
        
        """
        parts = ascore.getElementsByClass(Part)
        new_score = Score()
        for apart in parts:
            pname = apart.partName
            if partnames is None or pname in partnames:
                tp = MusicUtils.adjust_accidentals(apart, preference, inPlace)
                new_score.append(tp)
        return new_score
    
    @staticmethod
    def get_transposition_intervals(keys:[key.Key], key2:key.Key=C_Major, key2_minor:key.Key=A_Minor) -> [interval.Interval]:
        intval = []
        p2 = None
        for akey in keys:
            k = akey
            if not isinstance(akey, key.Key):
                k = akey.asKey()
            p1 = k.getPitches()[0]
            if k.mode == 'major':
                p2 = key2.getPitches()[0]
            else:
                p2 = key2_minor.getPitches()[0]
                
            interval_p1p2 = interval.Interval(noteStart = p1,noteEnd = p2)
            if abs(interval_p1p2.semitones) > 6:
                interval_p1p2 = interval_p1p2.complement
            if MusicUtils.verbose > 1:
                print('transposition key pitches {}, {} intervals p1p2 {}'.format(p1.nameWithOctave,p2.nameWithOctave, interval_p1p2.semitones))
            intval.append(interval_p1p2)
        return intval
    
    @staticmethod
    def transpose_part(apart:Part, target_key=C_Major, target_key_minor=A_Minor, \
                       instruments:Instruments=None, inPlace=False, adjustAccidentals=True) -> Part:
        """Transpose the notes in a given Part
            Args:
                apart : a stream.Part instance
                target_key : the transposition Key. Default is C-Major if not specified
                target_key_minor : the transposition Key if original key is minor mode. Default is A-Minor if not specified
                instruments - music.Instruments instance, if not None Part ranges are enforced after the transposition
                inPlace - if True, transpose in place. Otherwise transpose a copy
                adjustAccidentals - if True (the default) adjust all accidentals according to the caller's preference
            Returns:
                A Part with the notes transposed as specified.
                The Part argument is not modified if inPlace is False.
        """
        key_sigs,measure_numbers = MusicUtils.get_keySignatures(apart)
        measure_numbers.append(len(apart))
        transposition_intervals = MusicUtils.get_transposition_intervals(key_sigs, key2=target_key, key2_minor=target_key_minor)
        transposed_part = apart
        if not inPlace:     # transpose a copy of the Part
            transposed_part = copy.deepcopy(apart)
            
        for i in range(len(transposition_intervals)):
            #
            # no need to transpose if the measures are already in the target_key
            # but still need to adjust accidentals
            #
            intval = transposition_intervals[i]
            if MusicUtils.verbose > 1:
                print(f"transposition interval: {intval}")
            part_measures = transposed_part.measures(measure_numbers[i], measure_numbers[i+1]-1)
            if intval.semitones != 0:
                part_measures.transpose(intval, inPlace=True)
            if adjustAccidentals:
                MusicUtils.adjust_accidentals(part_measures)   # use defaults for preference and inPlace (True)
        
        if instruments is not None:
            #
            # check if notes are in range for Instrument and adjust if necessary
            #
            for anote in transposed_part.flat.getElementsByClass('Note'):
                instruments.adjust_to_range(apart.partName, anote, inPlace=True)
        
        return transposed_part
    
    @staticmethod
    def transpose_score(ascore:Score, target_key=C_Major, target_key_minor=A_Minor, partnames:[str]=None, \
                        instruments:Instruments=None, inPlace=False, adjustAccidentals=True) -> Score:
        """Transpose an entire Score to a new Key
        
            Args:
                ascore - the Score to transpose
                target_key - the new Key, default is C-Major
                partnames - part names to include (and transpose)
                instruments - music.Instruments instance, if not None Part ranges are enforced after the transposition
                inplace - if True, transpose in place. Otherwise transpose a copy
            Returns:
                a transposed Score. The original is not modified if inPlace is False
        
        """
        parts = ascore.getElementsByClass(Part)
        new_score = Score()
        for p in parts:
            pname = p.partName
            if partnames is None or pname in partnames:
                if MusicUtils.verbose > 1:
                    print('transpose {}'.format(pname))
                tp = MusicUtils.transpose_part(p, target_key=target_key, target_key_minor=target_key_minor, \
                                               instruments=instruments, inPlace=inPlace, adjustAccidentals=adjustAccidentals)
                new_score.append(tp)
        return new_score
    
    @staticmethod
    def extend_parts(ascore:Score, padlen:duration.Duration, time_signature = None):
        '''
        Extend each Part in a Score to padlen (total duration in quarterLengths)
        by adding rests.
        '''
        measureLen = 4.0
        if time_signature is not None:
            measureLen = time_signature.numerator * time_signature.beatDuration.quarterLength
        
        parts = ascore.getElementsByClass(Part)
        for p in parts:
            plen = p.duration.quarterLength
            padding = math.ceil(padlen.quarterLength/measureLen) * measureLen - plen
            if MusicUtils.verbose > 0:
                print(f'{p.partName} duration: {plen}')
                print(f' extend {p.partName} by {padding} quarterLengths')
            if padding > 0:
                r = note.Rest()
                r.duration.quarterLength = padding
                p.append(r)

    @staticmethod
    def rearrange(counts_df:pd.DataFrame, sortvalues=True) -> (pd.DataFrame, pd.DataFrame) :
        '''Restructure a counts_df DataFrame for plotting
        
            Args:
                counts_df - a counts DataFrame, as created by NoteCollector.
            Returns:
                2 new DataFrames structured as follows:
                pitchCounts_df has columns 'pitch', 'next_pitch', 'count'
                pitches_counts_df has columns 'pitch-next_pitch', 'count' (pitch, next_pitch concatenated & comma-separated into a single key
            Both DataFrames are sorted by pitch,next_pitch.
            
        '''
        pitchCounts_df = pd.DataFrame(columns = ['pitch', 'next_pitch', 'count'])
        pitches_counts_df = pd.DataFrame(columns = ['pitch-next_pitch', 'count'])
        pitchCounts_df
        pitches = list(counts_df['KEY'])
        for rownum in range(len(counts_df)):
            row = counts_df.iloc[rownum]
            from_pitch = row['KEY']
            for to_pitch in pitches:
                count = int(row[to_pitch])
                nrow = {'pitch':from_pitch, 'next_pitch':to_pitch, 'count':count}
                pcrow = {'pitch-next_pitch' : f"{from_pitch},{to_pitch}", 'count':count}
                pitchCounts_df = pitchCounts_df.append(nrow, ignore_index=True)
                pitches_counts_df = pitches_counts_df.append(pcrow, ignore_index=True)
        if sortvalues:
            pitchCounts_df.sort_values(by=['pitch', 'next_pitch'], ascending=[True, True], inplace=True)
            pitches_counts_df.sort_values(by=['pitch-next_pitch', 'count'], ascending=[True, True], inplace=True)
        return pitchCounts_df, pitches_counts_df

    @staticmethod
    def pivot_data(counts_df:pd.DataFrame, from_column_name='note1', to_column_name='note2'):
        ''' Rearrange the count data to use in a heat map
        
             Args:
                counts_df - a counts DataFrame, as created by NoteCollector 
        '''
        cols = counts_df.columns.sort_values()[:-1]
        note1 = []
        note2 = []
        count = []
        for rnum in range(len(counts_df['KEY'])):
            for n2 in cols:
                note1.append(counts_df.iloc[rnum]['KEY'])
                note2.append(n2)
                count.append(int(counts_df.iloc[rnum][n2]))
    
        data = {from_column_name:note1, to_column_name:note2, 'count':count}
        notes_df_raw = pd.DataFrame(data=data)
        notes_df = notes_df_raw.pivot(from_column_name,to_column_name,'count')

        return notes_df,notes_df_raw
    
    @staticmethod
    def plot_heat_map(counts_df:pd.DataFrame, fmt='d', linewidths=0.5, linecolor='LightGreen', cmap='viridis', annot=False):
        '''Plot counts DataFrame as a HeatMap. Note1 is on the Y-axis, Note2 (the note following Note1) on the X-axis
        
            Args:
                counts_df - a counts DataFrame, as created by NoteCollector
                fmt - String formatting code to use when adding annotations, default is 'd'
                linewidths:float  - line width to separate cells, default is 0.5
                linecolor - default is 'LightGreen'
                annot:bool or rectangular dataset, optional, default is False
                cmap:matplotlib colormap name or object, or list of colors - color map, default is 'viridis'
            
            For color maps see https://matplotlib.org/stable/gallery/color/colormap_reference.html
        '''
        nc_df = MusicUtils.pivot_data(counts_df)
        plt.figure(figsize=(12,8))
        sns.heatmap(nc_df[0], annot=annot, fmt=fmt, linewidths=linewidths, linecolor=linecolor, cmap=cmap)
        
    @staticmethod
    def plot_pitch_counts(counts_df:pd.DataFrame, sortvalues=True, style='whitegrid', kind='scatter', height=7, rot=-0.2, start=0.0):
        '''Create a Seaborn relational plot of pitch vs. next_pitch from a counts_df DataFrame
        
        Args:
            counts_df - a counts DataFrame, as created by NoteCollector
            style - plot style, default style is 'whitegrid'
            kind - kind of plot, can be 'line' or 'scatter' (the default) Would not recommend 'line'.
            height:int - the plot height, default is 7
            start:float, 0 <= start <= 3   The hue at the start of the helix. Default is 0.
            rot:float - Rotations around the hue wheel over the range of the palette. Default is -0.2
            
        The counts_df DataFrame is created by the NoteCollector in any of the modes: ap, apc, dp, dpc, or sd
        The default palette is cubehelix_palette. This produces a colormap with linearly-decreasing (or increasing) brightness.
        The size and hue of each plot point is determined by the count of (pitch, next_pitch)
        
        See https://seaborn.pydata.org/generated/seaborn.cubehelix_palette.html#seaborn.cubehelix_palette
        for color palette details.
        '''
        pitchCounts_df, pitches_counts_df = MusicUtils.rearrange(counts_df, sortvalues)
        cmap = sns.cubehelix_palette( rot=rot, start=start, as_cmap=True)
        sns.set_theme(style=style)
        g = sns.relplot(data=pitchCounts_df, x='pitch',y='next_pitch', height=height, \
                        size='count', hue='count', palette=cmap, kind=kind, sizes=(10,400))
        g.ax.xaxis.grid(True, "minor", linewidth=.25)
        g.ax.yaxis.grid(True, "minor", linewidth=.25)

if __name__ == '__main__':
    print(MusicUtils.__doc__)



import math

class Geometry(object):
    
    @staticmethod    
    def polygon_points(nsides = 24, radius = 400):
        theta = math.radians(360/nsides)
        points = []
        for i in range(nsides + 1):
            p = ( round(radius * math.cos(i * theta), 4), round(radius * math.sin(i * theta), 4) )
            points.append(p)
        return points

    @staticmethod    
    def points_dist(p1:(float,float), p2:(float,float)) -> float:
        '''Calculates the distance between two points.
        
        '''
        d = math.sqrt( (p2[0]-p1[0])**2 + (p2[1]-p1[1])**2 )
        return d

    @staticmethod    
    def segment_points(p1:(float,float), p2:(float,float), nsegments=4):
        points = []
        #
        # add in the points between p1 and p2
        #
        xd = (p2[0] - p1[0])/nsegments
        yd = (p2[1] - p1[1])/nsegments
        for n in range(nsegments):
            p = (p1[0] + n*xd, p1[1] + n*yd)
            points.append(p)
        points.append(p2)
        return points

    @staticmethod
    def polygon_segment_points(nsides = 12, radius = 400, nsegments = 4):
        """Get all segment points for a polygon.
            Arguments:
                nsides - polygon number of sides, default is 12
                radius - polygon radius in units, default is 400
                nsegments - number of segments to divide each pair of points, default is 4
            Returns:
                a dict having the polygon segment number (0 to nsegments-1) as the key, and the segment points as the value
            Segments are numbered counter-clockwise order with the first segment coordinate at x=nradius, y=0

        """
        points = Geometry.polygon_points(nsides=nsides, radius=radius)
        seg_points = {}
        for i in range(nsides):
            sp = Geometry.segment_points(points[i], points[i+1], nsegments)
            key = i # (points[i], points[i+1])
            seg_points[key] = sp
        return seg_points



'''
Created on Apr 26, 2021

@author: don_bacon
 1D avg: 0.3336056415465781     min distance: 6.953338937965015e-08    max distance 0.9991733125913103
 2D avg: 0.5218940528195612     min distance: 0.0002963836942284597    max distance 1.386732445768972
 3D avg: 0.6623875731449357     min distance: 0.006303274998869993     max distance 1.6275309823122912
 4D avg: 0.7786016428780859     min distance: 0.02219800547536187      max distance 1.817787360261562
 5D avg: 0.8795563100273328     min distance: 0.04582378971554501      max distance 1.9186141790261342
'''
import argparse, random, math, numpy as np

class Geo_test(object):
    
    def __init__(self, dims, iterations, trials):
        self.count = 0
        self.trial_count = 0
        self.dimensions = dims
        self.trials = trials
        self.iterations = iterations
        random.seed()
        self.average = 0.0
        self.verbose = 0
        self.max_dist = 0.0
        self.min_dist = 99.0
        self.std_normal = False
    
    def get_point(self):
        if self.std_normal:
            return np.random.randn(self.dimensions)
        else:
            return np.random.random(size=2)
    
    def distance_metric(self, point1, point2):  # 2 points in n-dimensions
        d = 0.0
        dim = len(point1)
        for i in range(0, dim):
            d = d + math.pow( point1[i] - point2[i], 2)
        d = math.sqrt(d)
        return d
        
    def run_trials(self):
        avg_sum = 0.0
        while self.trial_count < self.trials:
            avg_sum = avg_sum + self.run_trial()
            self.trial_count = self.trial_count + 1
            # random.seed()
        self.average = avg_sum / self.trials
        return self.average
        
    def run_trial(self):
        avg = 0.0
        if(self.dimensions == 1):
            avg = self.run_1d_trial()
        else:
            avg = self.run_nd_trial()
            
        if(self.verbose > 0):
            print("{}\t{}".format(self.trial_count, avg))
        return avg
    
    def run_nd_trial(self):
        dsum = 0.0
        for self.count in range(self.iterations):
            p1 = self.get_point()
            p2 = self.get_point()
            d = self.distance_metric(p1,p2)
            if d < self.min_dist:
                self.min_dist = d
            if d > self.max_dist:
                self.max_dist = d
            dsum = dsum+d
        return dsum / self.count
    
    def run_1d_trial(self):
        dsum = 0.0
        for self.count in range(self.iterations):
            p1=random.random()
            p2=random.random()
            d = math.fabs(p2-p1)
            if d < self.min_dist:
                self.min_dist = d
            if d > self.max_dist:
                self.max_dist = d
            dsum = dsum+d
        return dsum / self.count

    
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Point distance average.')
    parser.add_argument("dimensions", help="number of dimensions: 1 to 3", type=int )
    parser.add_argument("iterations", help="number of iterations per trial", type=int)
    parser.add_argument("--trials", help="number of trials", type=int, default = 1 )
    parser.add_argument('--verbose', '-v', action='count', default=0, help='verbose output level')
    parser.add_argument('--std', action='store_true', help="Use standard normal distribution")

    args = parser.parse_args()
    
    dimensions = args.dimensions
    trials = args.trials
    iterations = args.iterations
    geo_test = Geo_test(dimensions, iterations, trials)
    geo_test.verbose = args.verbose
    geo_test.std_normal = args.std
    avg = geo_test.run_trials()
    print("avg: {}\t min distance: {}\t max distance {}".format(avg, geo_test.min_dist, geo_test.max_dist))
    
    
    


'''
Created on Mar 3, 2021

@author: don_bacon
'''
import json
from music.scale import Scale

class Chord(object):

    def __init__(self, raw_data={}):
        '''
        Constructor
        '''
        # print("raw data: {}".format(raw_data))
        if(type(raw_data) is dict):
            self.chord = raw_data
        else:
            self.chord = json.loads(raw_data)
    
        self.name = self.chord['name']
        self.formula = self.chord['formula']
        self.intervals = self.chord['intervals']
        self.notes = self.chord['spelling'].split(' ')
        self.symbols = self.chord['symbols']
        self.size = len(self.formula)
        _scale = {}
        _scale['formula'] = self.formula
        _scale['name'] = self.name
        self.scale = Scale(_scale)
        



# ------------------------------------------------------------------------------
# Name:          keys.py
# Purpose:       Encapsulate music keys.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2021 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

import pandas as pd

class Keys(object):
    """Defines Keys from keys.json as a DataFrame
    
    """

    def __init__(self, verbose=0, resource_folder ="/Compile/dwbzen/resources/music"):
        
        self.resource_folder = resource_folder
        self.keys_file = resource_folder + "/keys.json"
        self.verbose = verbose
        self.keys_pd = pd.read_json(self.keys_file, orient="index")
        self.keys_pd.fillna(value={'parallelKey':'0'}, inplace=True)

if __name__ == '__main__':
    print(Keys.__doc__)
    keys = Keys()
    print(keys.keys_pd)




# ------------------------------------------------------------------------------
# Name:          instruments.py
# Purpose:       Instrument utilities.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2021 Donald Bacon

# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from music21 import instrument, Music21Object
from music21 import note, clef, interval

import pandas as pd
import importlib

class Instruments(object):
    """Encapsulates instrument information for the instruments supported by this framework:
       * music21 module
       * associated music21 class and class instance
       * range of the instrument - (low,high) as pitches
       * pitchRange - (low,high) as pitch pitch space attribute
       * Clef to use
       * transposeDiatonic
       * transposeChromatic
       * transposeText

    Instrument information maintained in resources/music/instruments.json
    Clef information in resources/music/clefs.json

    """
    
    instrument_names=[ 'Alto', 'Bass', 'Bassoon', 'Clarinet', 'Flute', 'Harpsichord', 'Koto', 'Oboe', 'Piano', 'PianoLH', 'PianoRH', 'Soprano', 'Tenor']
    clef_names=['Soprano', 'Alto', 'Tenor', 'Bass', 'Treble', 'Treble8va', 'Treble8vb', 'Bass8va', 'Bass8vb', 'C', 'F', 'G']
    
    def __init__(self, verbose=0, resource_folder ="/Compile/dwbzen/resources/music"):
        self.resource_folder = resource_folder
        self.verbose = verbose
        self.__create_instrument_classes()
        self.__create_clef_classes()

    @staticmethod
    def __create_instance(row:pd.Series) -> Music21Object:
        """Create an instance of a given class.
        
            Args:
                row - a pd.Series that has ''module' and 'class' members
            Returns:
                a module.class instance
        
        """
        module = row['module']
        class_name = row['class']
        my_module = importlib.import_module(module)
        MyClass = getattr(my_module, class_name)
        instance = MyClass()
        return instance
    
    @staticmethod
    def __define_pitch_range(row:pd.Series):
        notes = row['range'].split(",")
        nlow = note.Note(notes[0])
        nhigh = note.Note(notes[1])
        rlow = int(nlow.pitch.ps)
        rhigh = int(nhigh.pitch.ps)
        return [rlow,rhigh]
    
    def __create_instrument_classes(self):
        self.instruments_pd = pd.read_json(self.resource_folder + "/instruments.json", orient="index")
        self.instruments_pd['instance'] = [Instruments.__create_instance(row[1]) for row in self.instruments_pd.iterrows()]
        self.instruments_pd['range_ps'] = [Instruments.__define_pitch_range(row[1]) for row in self.instruments_pd.iterrows()]
        self.instruments_pd.fillna(value={'transposeDiatonic':0}, inplace=True)
        self.instruments_pd.fillna(value={'transposeChromatic':0}, inplace=True)
        self.instruments_pd.fillna(value={'transposeText':'Non-transposing'}, inplace=True)
    
    def __create_clef_classes(self):
        self.clefs_pd = pd.read_json(self.resource_folder + "/clefs.json", orient="index")
        self.clefs_pd['instance'] = [Instruments.__create_instance(row[1]) for row in self.clefs_pd.iterrows()]

    def get_clef(self, clef_name) -> clef.Clef:
        instance = None
        if clef_name in self.clefs_pd.index:
            instance = self.clefs_pd.loc[clef_name]['instance']
        return instance
    
    def get_instrument_clef(self, instrument_name) -> clef.Clef:
        instance = None
        if instrument_name in self.instruments_pd.index:
            clef_name = self.instruments_pd.loc[instrument_name]['clef']
            instance = self.get_clef(clef_name)
        return instance
    
    def get_instrument(self, instrument_name) -> instrument.Instrument:
        instance = None
        if instrument_name in self.instruments_pd.index:
            instance = self.instruments_pd.loc[instrument_name]['instance']
        return instance
    
    def is_in_range(self, instrument_name:str, note:note.Note) -> bool:
        """Return True if a Note is in the rage of a given Instrument, False otherwise
        
        """
        noteps = note.pitch.ps
        rng = self.instruments_pd.loc[instrument_name]['range_ps']
        inrange = (noteps >= rng[0] and noteps <= rng[1])
        return inrange
    
    def check_range(self, instrument_name:str, note:note.Note) -> int:
        """Returns the number of steps a Note is for a given Instrument, 0 otherwise.
        
            Args:
                instrument_name - the name of a supported Instrument
                note - the Note to test
            Returns:
                0 if in range
                if not in range, the #steps (semitones) the note is beyond the range.
                If <0, the Note is below by that number of steps
                If >0 the note is above by that number of steps
            Note:
                If note.octave is None, music21 assumes octave 4 for all notes.
                So 'C' is treated as 'C4' with a pitch.ps value of 60.
                and 'B' is treated as 'B4' with a pitch.ps value of 71.
        
        """
        steps_out_of_range = 0
        noteps = note.pitch.ps
        rng = self.instruments_pd.loc[instrument_name]['range_ps']
        if noteps < rng[0]:
            steps_out_of_range = noteps - rng[0]
        elif noteps > rng[1]:
            steps_out_of_range = noteps - rng[1]
        return int(steps_out_of_range)
    
    def adjust_to_range(self, instrument_name:str, anote:note.Note, inPlace=False) -> note.Note:
        """Adjust the pitch of a Note if out of range.
            If below the instrument range, the note is transposed up by the number of octaves needed to be back in range.
            If above the instrument range, the note is transposed down by the number of octaves needed to be back in range.
        
            Args:
                instrument_name - the name of a supported Instrument
                anote - the Note to test and possibly transpose
                inPlace - if True anote is transposed in place, else a new Note is returned and the original unchanged.
            Returns:
                a new note that is in range. If inPlace is True, the note provided is transposed in place.
                The original note is returned unchanged if already in range.
        
        """
        steps_out_of_range = self.check_range(instrument_name, anote)
        #
        # if the note is out of range for this part (instrument)
        # transpose up an octave is below the range, down an octave if above
        if steps_out_of_range != 0:
            if self.verbose > 1:
                print(f"newnote: {anote.nameWithOctave} out of range for {instrument_name}, by {steps_out_of_range} steps")
            
            noctaves = 1 + ( (abs(steps_out_of_range)-1) // 12)
            semitones = 12 * noctaves
            if steps_out_of_range <0:
                tintval = interval.Interval(semitones)
            elif steps_out_of_range > 0:
                tintval = interval.Interval(-semitones)
            else:
                tintval = interval.Interval(0)
            #
            # transpose does not return a value if inPlace == True
            #
            newnote = anote.transpose(tintval, inPlace=inPlace)
        else:
            newnote = anote

        return newnote
    
if __name__ == '__main__':
    print(Instruments.__doc__)
    instruments = Instruments()
    if instruments.verbose > 0:
        print(instruments.instruments_pd)
        print(instruments.clefs_pd)
    print(instruments.instruments_pd.loc['Flute'])
    
    
    


from setuptools import find_packages
from setuptools import setup
import os

ROOT_DIR = os.path.dirname(__file__)
SOURCE_DIR = os.path.join(ROOT_DIR)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requirements = [
    'music21 >=7.1.0',
    'pandas >=1.4.3'
]

setup(
    name="dwbzen",
    version="0.1.2",
    author="Don Bacon",
    author_email="dwbzen@gmail.com",
    description="Music and Math package including MarkovChain generation from text and music source, and production",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dwbzen/dwbzen",
    packages=find_packages(exclude=["test",]),
    data_files=[('resources', ['resources/music/allChordFormulas.json','resources/music/clefs.json', 'resources/music/common_scaleFormulas.json',
                  'resources/music/commonScaleFormulas.json','resources/music/instruments.json','resources/music/keys.json','resources/music/substitution_rules.json',
                  'resources/pos/2of12pos.txt', 'resources/pos/3eslpos.txt','resources/pos/5deskpos.txt',
                  'resources/text/drugBrandNames.txt'
                    ])],   
    install_requires=requirements,
    license="MIT",
    namespace_packages=[],
    include_package_data=False,
    classifiers=[
        "Development Status :: 3 - Alpha",
		"Intended Audience :: Developers",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
	    "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires='>=3.8',
    zip_safe=False,
    keywords="markov chain text music",
    maintainer='Donald Bacon',
    maintainer_email='dwbzen@gmail.com'
)



# ------------------------------------------------------------------------------
# Name:          sentenceProducer.py
# Purpose:       SentenceProducer class.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.producer import Producer
from common.textParser import TextParser
import random

class SentenceProducer(Producer):
    
    def __init__(self, state_size, markovChain, source_file, min_size=0, max_size=0, num=10, verbose=0, rand_seed=42):
        super().__init__(state_size, markovChain, source_file, min_size, max_size, num, verbose, rand_seed)
        self.list=False
        self.postprocessing = None
        self.initial=False
        self.pos=False

        self.terminal_characters = '.?!'   # needs to match the WordCollector terminal_characters
        self.output_format = 'TC'          # title case, can also do UC = upper case, LC = lower case

        self._set_keys()
    
    def __str__(self):
        return f"SentenceProducer order={self.order} verbose={self.verbose}, source={self.source}, min/max={self.min_size},{self.max_size} seed={self.seed}, num={self.num}"

    def _set_keys(self):
        prefix = TextParser._prefix
        self._initial_keys_df = self.counts_df[ [x.startswith(prefix) for x in self.counts_df['key']] ]
        #
        # initiial_keys and keys are both sets and so the values are unique
        #
        self._initial_keys = list(set(self._initial_keys_df['key'].values.tolist()))
        self._keys = list(set(self.counts_df['key'].values.tolist()))
        if self.verbose > 0:
            print(f'total number of keys: {len(self._keys)}')
            print(f'number of initial keys: {len(self._initial_keys)}')

    def get_seed(self, aseed=None):
        if self.seed is None or aseed is None:
            # need to pick one at random
            if self.initial:
                # pick a seed that starts a sentence (i.e. first character is a space)
                theseed = self._initial_keys[random.randint(0, len(self._initial_keys)-1)]
            else:
                # pick any old seed from the chain's index
                theseed = self._keys[random.randint(0, len(self._keys)-1)]
        else:
            theseed = aseed
        if self.verbose > 1:
            print(f'get seed: {theseed}')
        return theseed
    
    def get_next_word(self, seed):
        """Gets the next word based on the seed argument

        Returns:
            both the next word and a new seed as a dict with keys 'new_seed' and 'next_token'
        """
        
        new_seed = None
        next_token = None
        if self.verbose > 2:
            print(f"get_next_word(seed): '{seed}'")

        if seed in self._keys:
            df = self.chain_df[self.chain_df['key']==seed]
            if len(df) == 0:  #  add to chain_df from counts_df
                df = self._add_key(seed)
    
            # given a probability, pick the first word in the row_probs (cumulative probabilities)
            # having a probability > the random probability
            # for example:
            prob = random.random()
            row = df[df['prob']>prob].iloc[0]
            next_token = row['word'].lstrip()
            p = row['prob']
            
            new_seed = ' '.join(((seed + ' ' + next_token).split())[1:self.order+1])
            if self.verbose > 1:
                print(f"random prob: {prob}, row prob: {p}, seed: '{seed}', next_token: '{next_token}', new_seed: '{new_seed}'")
                
        return dict([ ('next_token', next_token), ('new_seed',new_seed)])

    def add_sentence_to_set(self, sentence_set, asentence):
        if self.output_format == 'TC':
            #
            # capitalize the first letter of the sentence
            #
            asentence = asentence[0].upper() + asentence[1:]
        elif self.output_format == 'LC':
            asentence = asentence.lower()
        elif self.output_format == 'UC':
            asentence = asentence.upper()
        else:
            pass
        sentence_set.add(asentence)
        
        if self.verbose > 1:
            print(f'added: {asentence}')
        if self.display_as_produced:
            print(asentence)
        return asentence
            
    def produce(self):
        """Produce a list of words using the MarkovChain probabilities
        
        """
        sentences = set()
        initial_seed = self.get_seed(self.seed)      # get the initial seed
        seed = initial_seed
        self.seed = seed
        if self.verbose > 0:
            print(f"initial seed: '{seed}'")
            
        for n in range(1, self.num+1):
            sentence = seed.strip()
            more_to_go = True
            #
            # add words to the sentence until it reaches a maximum length
            # or the next word is a terminator
            #
            nwords = self.order      # number of words in this sentence so far
            while more_to_go:
                nc_dict = self.get_next_word(seed)
                ntoken = nc_dict['next_token']
                nseed = nc_dict['new_seed']
                if self.verbose > 2:
                    print(f'ntoken: "{ntoken}"')
                    
                if ntoken is not None:
                    if ntoken in self.terminal_characters:
                        sentence = self.add_sentence_to_set(sentences, sentence)
                        more_to_go = False
                        seed = self.get_next_seed()
                        nwords = self.order
                    else:
                        sentence = f'{sentence} {ntoken.lstrip()}'
                        nwords+=1
                        if nwords >= self.max_size:
                            sentence = self.add_sentence_to_set(sentences, sentence + '.')
                            more_to_go = False
                            seed = self.get_next_seed()    # pick a new seed
                            nwords = self.order
                        else:
                            more_to_go = True
                            seed = nseed
                else:  
                    # this can happen when the collector is in processing sentences mode
                    # add the sentence as is and get a new seed
                    sentence = self.add_sentence_to_set(sentences, sentence + '.')
                    seed = self.get_seed()
                    nwords = self.order
                    more_to_go = False
                
        return sentences

if __name__ == '__main__':
    print(SentenceProducer.__doc__)




from music21 import stream, key, note
import pandas as pd

class MusicScale(object):
    def __init__(self,  resource_folder ="/Compile/dwbzen/resources/music", \
                 scale_name='Major', root_note=note.Note('C4'), key=key.Key('C')):
        
        self.scales_df = pd.read_json(resource_folder + "/commonScaleFormulas.json", orient='records').transpose()
        self.scale = self.scales_df.loc[scale_name]
        self.root = root_note
        self.scale_name = scale_name
        self.formula = self.scale['formula']
        self.key = key
        self.notes_stream = None
        self.scale_notes = self.get_scale_notes(root_note)
        self.scale_notes_names = [x.nameWithOctave for x in self.scale_notes]
        self.range_notes = self.get_range_notes(root_note)
        self.range_notes_names = [x.nameWithOctave for x in self.range_notes]
        
    def get_scale_notes(self, start_note=None):
        '''The notes of the configured scale spanning a single octave, for example 'C4' to 'C5'
            Note that the top note is the same as the bottom note but an octave higher.
            Returns: a [Note]
        '''
        ns = stream.Stream()
        ns.append(self.key)
        if start_note is None:
            n = self.root
        else:
            n = start_note
        ns.append(n)
        for i in range(len(self.formula)):
            n = n.transpose(self.formula[i], inPlace=False)
            ns.append(n)
        return [x for x in ns.notes]
    
    def get_range_notes(self, start_note=None, start_octave=0, end_octave=8):
        '''For example, with default arguments notes (Major scale) are C0,D0,...B7,C8
        
        '''
        if start_note is None:
            n = self.root
            start_note = self.root
        else:
            n = start_note
        notes_list = []
        for octave in range(start_octave, end_octave):
            for n in self.scale_notes[:-1]:
                new_note = note.Note(n.name+str(octave))
                notes_list.append(new_note)
        new_note = note.Note(start_note.name+str(end_octave))
        notes_list.append(new_note)
        return notes_list
    
    def get_note(self, scale_degrees:int=0, from_note=None) -> note.Note:
        '''Gets a new Note that is the given scale_degrees away from a given Note.
            Arguments:
                scale_degrees - any integer, the sign dictates the direction - up or down
                from_note - the starting note, default is the scale root note
            Returns: a new Note
            Note if from_note is not present in the full range of notes (self.range_notes), the from_note is returned.
        '''
        if from_note is None:
            from_note = self.root

        if from_note.nameWithOctave in self.range_notes_names:
            ind = self.range_notes_names.index(from_note.nameWithOctave) + scale_degrees
            return self.range_notes[ind]
        else:
            return from_note




from __future__ import absolute_import  # multi-line and relative/absolute imports

from .__version__ import __title__, __description__, __version__
from .__version__ import __author__, __author_email__, __license__
from .__version__ import __copyright__

__all__ = [
    'characterCollector',
    'characterCollectorRunner',
    'collector',
    'collectorProducer',
    'environment',
    'geometry',
    'markovChain',
    'producer',
    'ruleSet',
    'sentenceProducer',
    'sentenceProducerRunner',
    'substitutionSystem',
    'textParser',
    'utils',
    'wordCollector',
    'wordCollectorRunner',
    'wordProducer',
    'wordProducerRunner'
]

from dwbzen.common.characterCollector import CharacterCollector
from dwbzen.common.characterCollectorRunner import CharacterCollectorRunner
from dwbzen.common.collector import Collector
from dwbzen.common.collectorProducer import CollectorProducer
from dwbzen.common.environment import Environment
from dwbzen.common.geometry import Geometry
from dwbzen.common.markovChain import MarkovChain
from dwbzen.common.producer import Producer
from dwbzen.common.ruleSet  import RuleSet
from dwbzen.common.sentenceProducer import SentenceProducer
from dwbzen.common.sentenceProducerRunner import SentenceProducerRunner
from dwbzen.common.substitutionSystem import SubstitutionSystem
from dwbzen.common.textParser import TextParser
from dwbzen.common.utils import Utils
from dwbzen.common.wordCollector import WordCollector
from dwbzen.common.wordCollectorRunner import WordCollectorRunner
from dwbzen.common.wordProducer import WordProducer
from dwbzen.common.wordProducerRunner  import WordProducerRunner

environmentGlobal = Environment('dwbzen')




'''
Created on Feb 22, 2021

@author: don_bacon
'''
import json
from music.keys import Keys
from music.song import Song
from music.scales import Scales

def keys_usage(keys_file):   
    with open(keys_file, "r") as read_file:
        raw_data = json.load(read_file)
    # print(raw_data)
    
    keys = Keys(raw_data)
    
    print('Key names: {}'.format(keys.key_names))
    print(keys.key['B-Major'])
    
def song_usage(song_file):
    with open(song_file, "r") as read_file:
        song_data = json.load(read_file)
    song = Song(song_data)
    df_notes = song.df_notes
    df_harmony = song.df_harmony
    
    # select notes for section = 'verse-1'
    df_verse1_notes = df_notes.loc[df_notes.loc[:, 'section'] == 'verse-1', :]
    print(df_verse1_notes.head())
    
    # select rows having specific chords
    print(df_harmony[df_harmony['chord'].isin(['Bb', 'F7'])])
    
def scales_usage(scales_file):
    _scales = Scales(json_file_name=scales_file)
    print("scales file: {}".format(_scales.scales_file))
    print("scale names: {}".format(_scales.scale_names))
    # select a specific scale
    print("Dorian b2: \n{}".format(_scales.scale['Dorian b2'].scale))
    # get the DataFrame
    scales_df = _scales.scales_df
    scales_df_blues = scales_df[scales_df['groups'].apply(lambda x:'blues' in x)]
    
    print('blues scales with 7 notes: \n{}'.format( scales_df_blues[scales_df_blues['size']==7][['name', 'formula_str']]))


if __name__ == '__main__':
    print('Sample usage of Keys and Song')
    keys_file = "../../resources/keys.json"
    song_file = "../../resources/songs/Penny Lane.json"
    scales_file = "../../resources/common_scaleFormulas.json"
    keys_usage(keys_file)
    song_usage(song_file)
    scales_usage(scales_file)

    
    


# ------------------------------------------------------------------------------
# Name:          scales.py
# Purpose:       Encapsulate music scales.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------
from music.scale import Scale
import argparse
import json
import pandas as pd


class Scales(object):
    """Scale formulas for common scales
    
    """
    scales_file = "/common_scaleFormulas.json"
    
    pitches_flats =  ['C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B', 'C' ]
    pitches_sharps = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B', 'C' ]
    pitches_mixed =  ['C', 'C#', 'D', 'Eb', 'E', 'F', 'F#', 'G', 'G#', 'A', 'Bb', 'B', 'C' ]
    
    def __init__(self, verbose=0, resource_folder ="/Compile/dwbzen/resources/music"):

        self.scales = {}
        self.scale = {}
        self.data = {}
        self.scale_names = []
        self.resource_folder = resource_folder
        self.scales_file = resource_folder + Scales.scales_file
        self.verbose = verbose

        with open(self.scales_file, "r") as read_file:
            self.data = json.load(read_file)

        self.create_dataFrame()
        self.get_scales_data()
        
    def format_cols(self, formula, fnumber):
        return '{}:{}'.format(formula,fnumber)
    
    def get_pitch_set(self, formula):
        """
        Creates a List of relative pitch indexes, and a list of pitches for a given formula
        Each element is the number of steps from the root of 'C'.
        For example given the scale formula  [1, 2, 1, 2, 1, 2, 2, 1] (name == 'Shostakovich respelled all sharp')
        the resulting pitch set is [0, 1, 3, 4, 6, 7, 9, 11, 12] 
        If the root Pitch is C, the corresponding pitches are ['C#', 'Eb', 'E', 'F#', 'G', 'A', 'B', 'C']
        """
        ps = [0]
        pitches = ['C']
        for i in range(0,len(formula)):
            ind = formula[i] + ps[i]
            ps.append(ind)
            pitches.append(self.pitches_mixed[ind])
        return (ps, pitches)
    
    def formula_number(self, formula):
        """
        The formula number of a chord/scale formula is an integer unique to that formula
        It is created from the pitch_set of the formula by successive left shifts
        """
        fnum = 0
        (pitch_set) = self.get_pitch_set(formula)
        for i in pitch_set[0]:
            shiftamt = i
            if i >= 12:
                shiftamt = 12-1
            fnum = fnum + (1 << shiftamt)
        return fnum

    def create_dataFrame(self):
        self.column_names = ['name', 'alternateNames', 'groups', 'formula', 'size']
        self.scales_df = pd.DataFrame(data=self.data['scales'], columns=self.column_names)
        #
        # alternateNames need to be a list so if a scale doesn't have any alternateNames
        # set it to a single item list consisting of the name
        #
        missing_series = self.scales_df[self.scales_df['alternateNames'].isnull()]['name'].apply(lambda x:[x])
        self.scales_df['alternateNames'].fillna(missing_series, inplace=True)
        
        # add the formula number for each scale
        self.scales_df['formula_number'] =  self.scales_df['formula'].apply(self.formula_number)
        
        # create a formula_str column which combines the formula number and formula into a single string
        self.scales_df['formula_str'] = self.scales_df[['formula_number','formula']].apply(lambda df: self.format_cols( df['formula_number'], df['formula']),axis=1)

        # self.scales_df.set_index('name')
    

    def get_scales_data(self):
        self.scales = self.data['scales']
        for s in self.scales:
            self.scale_names.append(s['name'])
            self.scale[s['name']] = Scale(s)

    def __iter__(self):
        """ Iterate over the scales
        
        """
        return self.scales.__iter__()
        
if __name__ == '__main__':
    _scales = Scales()
    scales_df = _scales.scales_df
    parser = argparse.ArgumentParser()
    parser.add_argument("-n","--name", help="Scale name", type=str, default=None)
    parser.add_argument("-g", "--group", help="Group name", type=str, default=None)
    args = parser.parse_args()
    
    if args.name is not None:
        _scale = _scales.scale[args.name]
        print('{} formula: {}'.format(args.name, _scale.formula))
        _notes = _scale.notes(root='A', accidental_preference='mixed')
        print('{} notes: {}'.format(args.name, _notes))
        print(str(_scale))
    if args.group is not None:
        group_scales = scales_df[scales_df['groups'].apply(lambda x:args.group in x)]
        print(group_scales)
    